#include <Python.h>

#include <errno.h>
#include <unistd.h>
typedef __uint32_t u32;	/* hack, so we may include kernel's ethtool.h */
typedef __uint16_t u16;	/* ditto */
typedef __uint8_t u8;	/* ditto */
typedef __uint64_t u64;        /* ditto */
#include <iwlib.h>
#include <linux/ethtool.h>
#include <linux/sockios.h>


#define _PATH_PROCNET_DEV "/proc/net/dev"

static PyObject *
get_active_devices (PyObject * self, PyObject * args)
{
  PyObject * list;
  int numreqs = 30;
  struct ifconf ifc;
  struct ifreq *ifr;
  int n;
  int skfd;
  
  /* SIOCGIFCONF currently seems to only work properly on AF_INET sockets
     (as of 2.1.128) */ 
  /* Open control socket. */
  skfd = socket(AF_INET, SOCK_DGRAM, 0);

  if(skfd < 0) {
    PyErr_SetString(PyExc_OSError, strerror(errno));
    return NULL;
  }
  
  ifc.ifc_buf = NULL;
  for (;;) {
    ifc.ifc_len = sizeof(struct ifreq) * numreqs;
    ifc.ifc_buf = realloc(ifc.ifc_buf, ifc.ifc_len);
    
    if (ioctl(skfd, SIOCGIFCONF, &ifc) < 0) {
      PyErr_SetString(PyExc_OSError, strerror(errno));
      free(ifc.ifc_buf);
      close(skfd);
      return NULL;
    }
    if (ifc.ifc_len == sizeof(struct ifreq) * numreqs) {
      /* assume it overflowed and try again */
      numreqs += 10;
      continue;
    }
    break;
  }
  
  list = PyList_New(0);
  ifr = ifc.ifc_req;
  for (n = 0; n < ifc.ifc_len; n += sizeof(struct ifreq)) {
    if (!(ioctl(skfd, SIOCGIFFLAGS, ifr) < 0))
      if (ifr->ifr_flags & IFF_UP)
	PyList_Append(list, PyString_FromString(ifr->ifr_name));
    ifr++;
  }

  free(ifc.ifc_buf);
  close(skfd);

  return list;
}

static PyObject *
get_devices (PyObject * self, PyObject * args)
{
  PyObject *list;
  FILE *fd;
  char buffer[256];
  char *ret;;

  list = PyList_New(0);
  
  fd = fopen(_PATH_PROCNET_DEV, "r");
  if (fd == NULL) {
    PyErr_SetString(PyExc_OSError, strerror(errno));
    return NULL;
  }
  /* skip over first two lines */
  ret = fgets(buffer, 256, fd); ret = fgets(buffer, 256, fd);
  while (!feof(fd)) {
    char *name = buffer;
    char *end = buffer;

    if (fgets(buffer, 256, fd) == NULL)
      break;
    /* find colon */
    while (end && *end != ':') end++;
    *end = 0; /* terminate where colon was */
    while (*name == ' ') name++; /* skip over leading whitespace if any */
    PyList_Append(list, PyString_FromString(name));
  }
  fclose(fd);
  return(list);
}


static PyObject *
get_hwaddress (PyObject * self, PyObject * args)
{
  struct ethtool_cmd ecmd;
  struct ifreq ifr;
  int fd, err;
  char *devname;
  char hwaddr[20];

  if (!PyArg_ParseTuple(args, "s", &devname)) {
    return NULL;
  }

  /* Setup our control structures. */
  memset(&ecmd, 0, sizeof(ecmd));
  memset(&ifr, 0, sizeof(ifr));
  strncpy(&ifr.ifr_name[0], devname, IFNAMSIZ);
  ifr.ifr_name[IFNAMSIZ-1] = 0;
  
  /* Open control socket. */
  fd = socket(AF_INET, SOCK_DGRAM, 0);
  if(fd < 0) {
    PyErr_SetString(PyExc_OSError, strerror(errno));
    return NULL;
  }
  
  /* Get current settings. */
  err = ioctl(fd, SIOCGIFHWADDR, &ifr);
  if(err < 0) {
    char buf[2048];
    int eno = errno;
    sprintf(buf, "[Errno %d] %s", eno, strerror(eno));
    PyErr_SetString(PyExc_IOError, buf);
    close(fd);
    return NULL;
  }

  close(fd);

  sprintf(hwaddr, "%02x:%02x:%02x:%02x:%02x:%02x",
	  (unsigned int)ifr.ifr_hwaddr.sa_data[0] % 256,
	  (unsigned int)ifr.ifr_hwaddr.sa_data[1] % 256,
	  (unsigned int)ifr.ifr_hwaddr.sa_data[2] % 256,
	  (unsigned int)ifr.ifr_hwaddr.sa_data[3] % 256,
	  (unsigned int)ifr.ifr_hwaddr.sa_data[4] % 256,
	  (unsigned int)ifr.ifr_hwaddr.sa_data[5] % 256);

  return PyString_FromString (hwaddr);
}


static  PyObject *
get_ipaddress (PyObject * self, PyObject * args)
{
  
  struct ethtool_cmd ecmd;
  struct ifreq ifr;
  int fd, err;
  char *devname;
  char ipaddr[20];

  if (!PyArg_ParseTuple(args, "s", &devname)) {
    return NULL;
  }

  /* Setup our control structures. */
  memset(&ecmd, 0, sizeof(ecmd));
  memset(&ifr, 0, sizeof(ifr));
  strncpy(&ifr.ifr_name[0], devname, IFNAMSIZ);
  ifr.ifr_name[IFNAMSIZ-1] = 0;

  /* Open control socket. */
  fd = socket(AF_INET, SOCK_DGRAM, 0);
  if(fd < 0) {
    PyErr_SetString(PyExc_OSError, strerror(errno));
    return NULL;
  }

    /* Get current settings. */
  err = ioctl(fd, SIOCGIFADDR, &ifr);
  if(err < 0) {
    char buf[2048];
    int eno = errno;
    sprintf(buf, "[Errno %d] %s", eno, strerror(eno));
    PyErr_SetString(PyExc_IOError, buf);
    close(fd);
    return NULL;
  }

  close(fd);


  sprintf(ipaddr, "%u.%u.%u.%u",
	  (unsigned int)ifr.ifr_addr.sa_data[2] % 256,
	  (unsigned int)ifr.ifr_addr.sa_data[3] % 256,
	  (unsigned int)ifr.ifr_addr.sa_data[4] % 256,
	  (unsigned int)ifr.ifr_addr.sa_data[5] % 256);

  return PyString_FromString (ipaddr);
}


static  PyObject *
get_netmask (PyObject * self, PyObject * args)
{
  
  struct ethtool_cmd ecmd;
  struct ifreq ifr;
  int fd, err;
  char *devname;
  char netmask[20];

  if (!PyArg_ParseTuple(args, "s", &devname)) {
    return NULL;
  }

  /* Setup our control structures. */
  memset(&ecmd, 0, sizeof(ecmd));
  memset(&ifr, 0, sizeof(ifr));
  strncpy(&ifr.ifr_name[0], devname, IFNAMSIZ);
  ifr.ifr_name[IFNAMSIZ-1] = 0;

  /* Open control socket. */
  fd = socket(AF_INET, SOCK_DGRAM, 0);
  if(fd < 0) {
    PyErr_SetString(PyExc_OSError, strerror(errno));
    return NULL;
  }

    /* Get current settings. */
  err = ioctl(fd, SIOCGIFNETMASK, &ifr);
  if(err < 0) {
    char buf[2048];
    int eno = errno;
    sprintf(buf, "[Errno %d] %s", eno, strerror(eno));
    PyErr_SetString(PyExc_IOError, buf);
    close(fd);
    return NULL;
  }

  close(fd);


  sprintf(netmask, "%u.%u.%u.%u",
	  (unsigned int)ifr.ifr_netmask.sa_data[2] % 256,
	  (unsigned int)ifr.ifr_netmask.sa_data[3] % 256,
	  (unsigned int)ifr.ifr_netmask.sa_data[4] % 256,
	  (unsigned int)ifr.ifr_netmask.sa_data[5] % 256);

  return PyString_FromString (netmask);
}





static  PyObject *
get_broadcast (PyObject * self, PyObject * args)
{
  
  struct ethtool_cmd ecmd;
  struct ifreq ifr;
  int fd, err;
  char *devname;
  char broadcast[20];

  if (!PyArg_ParseTuple(args, "s", &devname)) {
    return NULL;
  }

  /* Setup our control structures. */
  memset(&ecmd, 0, sizeof(ecmd));
  memset(&ifr, 0, sizeof(ifr));
  strncpy(&ifr.ifr_name[0], devname, IFNAMSIZ);
  ifr.ifr_name[IFNAMSIZ-1] = 0;

  /* Open control socket. */
  fd = socket(AF_INET, SOCK_DGRAM, 0);
  if(fd < 0) {
    PyErr_SetString(PyExc_OSError, strerror(errno));
    return NULL;
  }

    /* Get current settings. */
  err = ioctl(fd, SIOCGIFBRDADDR, &ifr);
  if(err < 0) {
    char buf[2048];
    int eno = errno;
    sprintf(buf, "[Errno %d] %s", eno, strerror(eno));
    PyErr_SetString(PyExc_IOError, buf);
    close(fd);
    return NULL;
  }

  close(fd);


  sprintf(broadcast, "%u.%u.%u.%u",
	  (unsigned int)ifr.ifr_broadaddr.sa_data[2] % 256,
	  (unsigned int)ifr.ifr_broadaddr.sa_data[3] % 256,
	  (unsigned int)ifr.ifr_broadaddr.sa_data[4] % 256,
	  (unsigned int)ifr.ifr_broadaddr.sa_data[5] % 256);

  return PyString_FromString (broadcast);
}



static PyObject *
get_module (PyObject * self, PyObject * args)
{
  struct ethtool_cmd ecmd;
  struct ifreq ifr;
  int fd, err;
  char buf[2048];
  char *devname;
    
  if (!PyArg_ParseTuple(args, "s", &devname)) {
    return NULL;
  }

  /* Setup our control structures. */
  memset(&ecmd, 0, sizeof(ecmd));
  memset(&ifr, 0, sizeof(ifr));
  strncpy(&ifr.ifr_name[0], devname, IFNAMSIZ);
  ifr.ifr_name[IFNAMSIZ-1] = 0;
  ifr.ifr_data = (caddr_t) &buf;
  ecmd.cmd = ETHTOOL_GDRVINFO;
  memcpy(&buf, &ecmd, sizeof(ecmd));
  
  /* Open control socket. */
  fd = socket(AF_INET, SOCK_DGRAM, 0);
  if(fd < 0) {
    PyErr_SetString(PyExc_OSError, strerror(errno));
    return NULL;
  }
  
  /* Get current settings. */
  err = ioctl(fd, SIOCETHTOOL, &ifr);

  if(err < 0) {  /* failed? */
    int eno = errno;
    FILE *file;
    int found = 0;
    char driver[100], dev[100];

    close(fd);

    /* before bailing, maybe it is a PCMCIA/PC Card? */
    file = fopen("/var/lib/pcmcia/stab", "r");
    if (file == NULL) {
      sprintf(buf, "[Errno %d] %s", eno, strerror(eno));
      PyErr_SetString(PyExc_IOError, buf);
      return NULL;
    }
    while (!feof(file)) {
      if (fgets(buf, 2048, file) == NULL)
	break;
      buf[2047] = '\0';
      if (strncmp(buf, "Socket", 6) != 0) {
	if (sscanf(buf, "%*d\t%*s\t%100s\t%*d\t%100s\n", driver, dev) > 0) {
	  driver[99] = '\0';
	  dev[99] = '\0';
	  if (strcmp(devname, dev) == 0) {
	    found = 1;
	    break;
	  }
	}
      }
    }
    fclose(file);
    if (!found) {
      sprintf(buf, "[Errno %d] %s", eno, strerror(eno));
      PyErr_SetString(PyExc_IOError, buf);
      return NULL;
    }
    else
      return PyString_FromString (driver);
  }

  close(fd);
  return PyString_FromString (((struct ethtool_drvinfo *)buf)->driver);
}

static PyObject *
get_businfo (PyObject * self, PyObject * args)
{
  struct ethtool_cmd ecmd;
  struct ifreq ifr;
  int fd, err;
  char buf[1024];
  char *devname;
    
  if (!PyArg_ParseTuple(args, "s", &devname)) {
    return NULL;
  }

  /* Setup our control structures. */
  memset(&ecmd, 0, sizeof(ecmd));
  memset(&ifr, 0, sizeof(ifr));
  strncpy(&ifr.ifr_name[0], devname, IFNAMSIZ);
  ifr.ifr_name[IFNAMSIZ-1] = 0;
  ifr.ifr_data = (caddr_t) &buf;
  ecmd.cmd = ETHTOOL_GDRVINFO;
  memcpy(&buf, &ecmd, sizeof(ecmd));
  
  /* Open control socket. */
  fd = socket(AF_INET, SOCK_DGRAM, 0);
  if(fd < 0) {
    PyErr_SetString(PyExc_OSError, strerror(errno));
    return NULL;
  }
  
  /* Get current settings. */
  err = ioctl(fd, SIOCETHTOOL, &ifr);

  if(err < 0) {  /* failed? */
    int eno = errno;
    close(fd);

    sprintf(buf, "[Errno %d] %s", eno, strerror(eno));
    PyErr_SetString(PyExc_IOError, buf);
    return NULL;
  }

  close(fd);
  return PyString_FromString (((struct ethtool_drvinfo *)buf)->bus_info);
}


/*------------------------------------------------------------------*/
/*
 * Get wireless informations & config from the device driver
 * We will call all the classical wireless ioctl on the driver through
 * the socket to know what is supported and to get the settings...
 */
static int
get_info(int			skfd,
	 char *			ifname,
	 struct wireless_info *	info)
{
  struct iwreq		wrq;

  memset((char *) info, 0, sizeof(struct wireless_info));

  /* Get wireless name */
  if(iw_get_ext(skfd, ifname, SIOCGIWNAME, &wrq) < 0)
    {
      /* If no wireless name : no wireless extensions */
      /* But let's check if the interface exists at all */
      struct ifreq ifr;

      strncpy(&ifr.ifr_name[0], ifname, IFNAMSIZ);
      ifr.ifr_name[IFNAMSIZ-1] = 0;
      if(ioctl(skfd, SIOCGIFFLAGS, &ifr) < 0)
	return(-ENODEV);
      else
	return(-ENOTSUP);
    }
  else {
    strncpy(info->b.name, wrq.u.name, sizeof(info->b.name));
    info->b.name[sizeof(info->b.name)-1] = 0;
  }

  /* Get ranges */
  if(iw_get_range_info(skfd, ifname, &(info->range)) >= 0)
    info->has_range = 1;

  /* Get network ID */
  if(iw_get_ext(skfd, ifname, SIOCGIWNWID, &wrq) >= 0)
    {
      info->b.has_nwid = 1;
      memcpy(&(info->b.nwid), &(wrq.u.nwid), sizeof(iwparam));
    }

  /* Get frequency / channel */
  if(iw_get_ext(skfd, ifname, SIOCGIWFREQ, &wrq) >= 0)
    {
      info->b.has_freq = 1;
      info->b.freq = iw_freq2float(&(wrq.u.freq));
    }

  /* Get sensitivity */
  if(iw_get_ext(skfd, ifname, SIOCGIWSENS, &wrq) >= 0)
    {
      info->has_sens = 1;
      memcpy(&(info->sens), &(wrq.u.sens), sizeof(iwparam));
    }

  /* Get encryption information */
  wrq.u.data.pointer = (caddr_t) info->b.key;
  wrq.u.data.length = IW_ENCODING_TOKEN_MAX;
  wrq.u.data.flags = 0;
  if(iw_get_ext(skfd, ifname, SIOCGIWENCODE, &wrq) >= 0)
    {
      info->b.has_key = 1;
      info->b.key_size = wrq.u.data.length;
      info->b.key_flags = wrq.u.data.flags;
    }

  /* Get ESSID */
  wrq.u.essid.pointer = (caddr_t) info->b.essid;
  wrq.u.essid.length = IW_ESSID_MAX_SIZE;
  wrq.u.essid.flags = 0;
  if(iw_get_ext(skfd, ifname, SIOCGIWESSID, &wrq) >= 0)
    {
      info->b.has_essid = 1;
      info->b.essid_on = wrq.u.data.flags;
    }

  /* Get AP address */
  if(iw_get_ext(skfd, ifname, SIOCGIWAP, &wrq) >= 0)
    {
      info->has_ap_addr = 1;
      memcpy(&(info->ap_addr), &(wrq.u.ap_addr), sizeof (sockaddr));
    }

  /* Get NickName */
  wrq.u.essid.pointer = (caddr_t) info->nickname;
  wrq.u.essid.length = IW_ESSID_MAX_SIZE;
  wrq.u.essid.flags = 0;
  if(iw_get_ext(skfd, ifname, SIOCGIWNICKN, &wrq) >= 0)
    if(wrq.u.data.length > 1)
      info->has_nickname = 1;

  /* Get bit rate */
  if(iw_get_ext(skfd, ifname, SIOCGIWRATE, &wrq) >= 0)
    {
      info->has_bitrate = 1;
      memcpy(&(info->bitrate), &(wrq.u.bitrate), sizeof(iwparam));
    }

  /* Get RTS threshold */
  if(iw_get_ext(skfd, ifname, SIOCGIWRTS, &wrq) >= 0)
    {
      info->has_rts = 1;
      memcpy(&(info->rts), &(wrq.u.rts), sizeof(iwparam));
    }

  /* Get fragmentation threshold */
  if(iw_get_ext(skfd, ifname, SIOCGIWFRAG, &wrq) >= 0)
    {
      info->has_frag = 1;
      memcpy(&(info->frag), &(wrq.u.frag), sizeof(iwparam));
    }

  /* Get operation mode */
  if(iw_get_ext(skfd, ifname, SIOCGIWMODE, &wrq) >= 0)
    {
      info->b.mode = wrq.u.mode;
      if((info->b.mode < IW_NUM_OPER_MODE) && (info->b.mode >= 0))
	info->b.has_mode = 1;
    }

  /* Get Power Management settings */
  wrq.u.power.flags = 0;
  if(iw_get_ext(skfd, ifname, SIOCGIWPOWER, &wrq) >= 0)
    {
      info->has_power = 1;
      memcpy(&(info->power), &(wrq.u.power), sizeof(iwparam));
    }

#if WIRELESS_EXT > 9
  /* Get Transmit Power */
  if(iw_get_ext(skfd, ifname, SIOCGIWTXPOW, &wrq) >= 0)
    {
      info->has_txpower = 1;
      memcpy(&(info->txpower), &(wrq.u.txpower), sizeof(iwparam));
    }
#endif

#if WIRELESS_EXT > 10
  /* Get retry limit/lifetime */
  if(iw_get_ext(skfd, ifname, SIOCGIWRETRY, &wrq) >= 0)
    {
      info->has_retry = 1;
      memcpy(&(info->retry), &(wrq.u.retry), sizeof(iwparam));
    }
#endif	/* WIRELESS_EXT > 10 */

  /* Get stats */
  if(iw_get_stats(skfd, ifname, &(info->stats), &(info->range), info->has_range) >= 0)
    {
      info->has_stats = 1;
    }

  return(0);
}

static PyObject *
get_iwconfig (PyObject * self, PyObject * args)
{
  char *devname;
  int skfd;		/* generic raw socket desc.	*/
  char buffer[1024];
  int eno = 0;
  struct wireless_info	info;
  PyObject * dict;
    
  if (!PyArg_ParseTuple(args, "s", &devname)) {
    return NULL;
  }

  /* Create a channel to the NET kernel. */
  if((skfd = iw_sockets_open()) < 0) {
    eno = errno;
    sprintf(buffer, "iw_sockets_open [Errno %d] %s", eno, strerror(eno));
    PyErr_SetString(PyExc_IOError, buffer);
    return NULL;
  }

  eno = get_info(skfd, devname, &info);

  close(skfd);

  if (eno < 0) {
    sprintf(buffer, "get_info [Errno %d] %s", eno, strerror(eno));
    PyErr_SetString(PyExc_IOError, buffer);
    return NULL;
  }

  dict = PyDict_New();
  if (info.b.has_mode)
    PyDict_SetItem(dict, PyString_FromString("Mode"), 
		   PyString_FromString(iw_operation_mode[info.b.mode]));
  if (info.b.essid_on) {
    PyDict_SetItem(dict, PyString_FromString("ESSID"), PyString_FromString(info.b.essid));
  } else {
    PyDict_SetItem(dict, PyString_FromString("ESSID"), PyString_FromString("Auto"));
  }
  if (info.b.has_nwid) {
    if(info.b.nwid.disabled)
      PyDict_SetItem(dict, PyString_FromString("NWID"), PyString_FromString("Auto"));
    else
      PyDict_SetItem(dict, PyString_FromString("NWID"), PyString_FromFormat("%X", info.b.nwid.value));
  }

  /* Display frequency / channel */
  if(info.b.has_freq)
    {
      iw_print_freq_value(buffer, sizeof(buffer), info.b.freq);
      PyDict_SetItem(dict, PyString_FromString("Frequency"), PyString_FromString(buffer));
    }

  /* Display the address of the current Access Point */
  if(info.has_ap_addr)
    {
      iw_pr_ether(buffer, (unsigned char *)info.ap_addr.sa_data);
      /* Oups ! No Access Point in Ad-Hoc mode */
      if((info.b.has_mode) && (info.b.mode == IW_MODE_ADHOC))
	PyDict_SetItem(dict, PyString_FromString("Cell"), PyString_FromString(buffer));
      else
	PyDict_SetItem(dict, PyString_FromString("Access Point"), PyString_FromString(buffer));
    }

  /* Display the currently used/set bit-rate */
  if(info.has_bitrate)
    {
      /* Display it */
      iw_print_bitrate(buffer, sizeof(buffer), info.bitrate.value);
      PyDict_SetItem(dict, PyString_FromString("BitRate"), PyString_FromString(buffer));
/*       printf("Bit Rate%c%s   ", (info.bitrate.fixed ? '=' : ':'), buffer); */
    }

  /* Display encryption information */
  /* Note : we display only the "current" key, use iwlist to list all keys */
  if(info.b.has_key)
    {
      if((info.b.key_flags & IW_ENCODE_DISABLED) || (info.b.key_size == 0))
	PyDict_SetItem(dict, PyString_FromString("Key"), PyString_FromString("off"));
      else
	{
	  /* Display the key */
	  iw_print_key(buffer, sizeof(buffer), info.b.key, info.b.key_size, info.b.key_flags);
	PyDict_SetItem(dict, PyString_FromString("Key"), PyString_FromString(buffer));

  	} 
    }

  return(dict);

}


static struct PyMethodDef PyEthModuleMethods[] = {
    	{ "get_module",
          (PyCFunction) get_module, METH_VARARGS, NULL },
    	{ "get_businfo",
          (PyCFunction) get_businfo, METH_VARARGS, NULL },
    	{ "get_hwaddr",
          (PyCFunction) get_hwaddress, METH_VARARGS, NULL },
	{ "get_ipaddr",
	  (PyCFunction) get_ipaddress, METH_VARARGS, NULL },
	{ "get_netmask",
	  (PyCFunction) get_netmask, METH_VARARGS, NULL },
	{ "get_broadcast",
	  (PyCFunction) get_broadcast, METH_VARARGS, NULL },
    	{ "get_iwconfig",
          (PyCFunction) get_iwconfig, METH_VARARGS, NULL },
    	{ "get_devices",
          (PyCFunction) get_devices, METH_VARARGS, NULL },
    	{ "get_active_devices",
          (PyCFunction) get_active_devices, METH_VARARGS, NULL },
	{ NULL, NULL, 0, NULL }	
};

void initethtool(void) {
  PyObject *m, *d;
  
  m = Py_InitModule("ethtool", PyEthModuleMethods);
  d = PyModule_GetDict(m);
}
